"""Middleware classes for using with the zookeeper library."""

from typing import Callable

from django.http.request import HttpRequest
from django.http.response import HttpResponse

from .connection import zookeeper_connection_manager


class ZookeeperConnectionMiddleware:

    """Runs the zookeeper connection manager while the request is being processed."""

    def __init__(self, get_response: Callable[[HttpRequest], HttpResponse]):
        """Store thr `get_response` function as an instance attribute."""
        self.get_response = get_response

    def __call__(self, request: HttpRequest) -> HttpResponse:
        """Generate the response in the managed context."""
        with zookeeper_connection_manager:
            return self.get_response(request)



"""An atomic version of the original migrate management command."""

from contextlib import contextmanager

from django.conf import settings
from django.core.exceptions import ImproperlyConfigured
from django.core.management.commands.migrate import Command as BaseCommand
from django.db import connection
from django.db.migrations.executor import MigrationExecutor

from zookeeper_locks.locks import Lock


@contextmanager
def nullcontext():
    yield


class Command(BaseCommand):

    """A migrate management command that executes in a critical section."""

    def __init__(self, lock=None, *args, **kwarg):
        self.lock = lock or Lock('migrations')
        super(Command, self).__init__(*args, **kwarg)

    def launched_with_defaults(self, **options):
        return (
            all(not options.get(opt) for opt in ['settings', 'pythonpath', 'app_label', 'migration_name', 'fake', 'fake_initial', 'run_syncdb'])
            and options.get('database') == 'default'
        )

    def handle(self, *args, **options):
        self.stdout.write(self.style.MIGRATE_HEADING("Potential data migration will be assisted by Zookeeper locks."))
        try:
            executor = MigrationExecutor(connection)
        except ImproperlyConfigured:
            # No databases are configured (or the dummy one)
            has_unapplied_migrations = True  # let the superclass decide what to do
        else:
            has_unapplied_migrations = bool(executor.migration_plan(executor.loader.graph.leaf_nodes()))

        if not self.launched_with_defaults(**options) or has_unapplied_migrations:
            if settings.ZOOKEEPER_HOSTS:
                lock_ctx = self.lock
            else:
                self.stdout.write(
                    self.style.NOTICE(
                        "No host has been defined in `settings.ZOOKEEPER_HOSTS` - Zookeeper locks will not protect current data migration process."
                    )
                )
                lock_ctx = nullcontext
            with lock_ctx():
                super(Command, self).handle(*args, **options)
        else:
            self.stdout.write(self.style.MIGRATE_LABEL("No migrations to apply."))



"""Utility for managing the Zookeeper connection for ZookeeperClient instances."""

import threading
from contextlib import ContextDecorator

from kazoo.client import KazooClient as ZookeeperClient
from kazoo.exceptions import ConnectionClosedError

from django.conf import settings


__all__ = ['zookeeper_connection_manager']


class ZookeeperConnectionManager(ContextDecorator):

    """A context manager / decorator for managing the Zookeeper connection."""

    # data holds objects shared among all instances of ZookeeperConnectionManager:
    #  - client - the ZookeeperClient instance
    #  - reference_counter - used to determine when to stop the client connection
    _data = threading.local()
    _data_lock = threading.Lock()

    def start_context(self, *args, **kwargs):
        """Increment the reference counter."""
        with self._data_lock:
            if not hasattr(self._data, 'reference_counter'):
                self._data.reference_counter = 0
            self._data.reference_counter += 1

    def stop_context(self, *args, **kwargs):
        """Decrement the reference counter and stop the connection while exiting the outermost context."""
        with self._data_lock:
            assert self.is_managed, 'Calling stop_context before start_context.'
            self._data.reference_counter -= 1
            if self._data.reference_counter == 0 and self.has_client:
                self._stop_connection()

    def get_client(self) -> ZookeeperClient:
        """Return a connected ZookeeperClient instance."""
        with self._data_lock:
            assert self.is_managed, 'Use the zookeeper_locks.connection.ZookeeperConnectionManager as a context manager or decorator.'
            if not self.has_client:
                self._start_connection()
            return self._data.client

    @property
    def is_managed(self) -> bool:
        """Determine if the managed context is active."""
        return getattr(self._data, 'reference_counter', 0) > 0

    @property
    def has_client(self) -> bool:
        """Determine if the client has been created."""
        return bool(getattr(self._data, 'client', None))

    def __enter__(self) -> 'ZookeeperConnectionManager':
        """Enter the context when the manager is used as a context manager or decorator."""
        self.start_context()
        return self

    def __exit__(self, exc_type, *exc) -> bool:
        """Exit the context when the manager is used as a context manager or decorator."""
        self.stop_context()
        if exc_type is ConnectionClosedError and self.has_client:
            self._data.client.restart()
        return False

    def _start_connection(self):
        """Create a ZookeeperClient instance and establish a connection."""
        self._data.client = self._create_client()
        self._data.client.start()

    def _stop_connection(self):
        """Stop the client connection and set the client attribute to None."""
        self._data.client.stop()
        self._data.client = None

    def _create_client(self) -> ZookeeperClient:
        return ZookeeperClient(
            hosts=','.join(settings.ZOOKEEPER_HOSTS),
        )


zookeeper_connection_manager = ZookeeperConnectionManager()



"""Multiprocessing hooks."""
from multiprocessing.util import Finalize

from zookeeper_locks.connection import zookeeper_connection_manager


def zookeeper_multiprocessing_connection_initializer():
    """Zookeeper multiprocessing connection initializer. Connection will not be closed on an raise of unhandled exception"""
    zookeeper_connection_manager.start_context()
    Finalize(zookeeper_connection_manager, zookeeper_connection_manager.stop_context, exitpriority=16)



"""Tools for creating distributed locks."""
import os
import threading
from contextlib import contextmanager
from functools import wraps
from logging import getLogger
from typing import (
    Any,
    Dict,
    Optional,
)

from kazoo.exceptions import LockTimeout as ZookeeperLockTimeout

from django.conf import settings
from django.core.exceptions import ImproperlyConfigured

from .connection import zookeeper_connection_manager


logger = getLogger(__name__)


class LockTimeout(Exception):

    """Indicates that lock has failed to acquire during the timeout period."""


class Locked(Exception):

    """Indicates that the lock couldn't be acquired immediately."""


class Lock:

    """Exclusive distributed lock backed with Zookeeper.

    Example:

    >>> lock = Lock('my-lock-{object_id}')
    >>>
    >>> with lock(object_id=123):
    >>>     print('so alone')
    >>>
    >>> try:
    >>>     with lock(object_id=123, timeout=9):
    >>>         print('so alone')
    >>> except LockTimeout:
    >>>     print('unable to lock after waiting for 9s')
    >>>
    >>> try:
    >>>     with lock(object_id=123, blocking=False):
    >>>         print('so alone')
    >>> except Locked:
    >>>     print('unable to lock immediately')
    """

    # map key to lock object id
    keys_registry: Dict[str, int] = dict()

    def __init__(self, key: str):
        """Register the given key and store it in the instance."""
        self.key = key
        self._register()
        self._locked_store = threading.local()

    def __del__(self):
        """Unregister the key when destructing a lock."""
        for key, value in self.keys_registry.items():
            if value == id(self):
                del self.keys_registry[key]
                break

    def _register(self):
        """Make sure the key has not been already used and add the key to the registry."""
        if self.key in self.keys_registry:
            raise ImproperlyConfigured('Attempt to register the same key twice: {}'.format(self.key))
        self.keys_registry[self.key] = id(self)

    @contextmanager
    def __call__(self, blocking: bool = True, timeout: Optional[float] = None, **key_params: Any):
        """Connect to Zookeeper and acquire the lock."""
        self._setup_locked_store()
        key_with_params = self._get_key_with_params(key_params)
        if key_with_params in self._get_locked_keys():
            yield
        else:
            with zookeeper_connection_manager:
                zk = zookeeper_connection_manager.get_client()
                lock = zk.Lock('/locks/{namespace}/{key}'.format(namespace=settings.ZOOKEEPER_APP_NAMESPACE, key=key_with_params))
                logger.info('Acquiring lock', extra={"namespace": settings.ZOOKEEPER_APP_NAMESPACE, "key": key_with_params})
                try:
                    acquired = lock.acquire(blocking=blocking, timeout=timeout)
                except ZookeeperLockTimeout as e:
                    raise LockTimeout('Timeout occurred while trying to acquire a blocking lock on {}'.format(key_with_params)) from e
                if acquired:
                    try:
                        self._add_locked_key(key_with_params)
                        yield
                    finally:
                        lock.release()
                        self._remove_locked_key(key_with_params)
                if not acquired:
                    raise Locked('Failed to acquire a non-blocking lock on {}'.format(key_with_params))

    def _get_key_with_params(self, key_params: Dict[str, Any]) -> str:
        """Format key with the given parameters."""
        return self.key.format(**key_params)

    def _get_locked_keys(self):
        """Returns Process Specific set of locked keys"""
        pid = os.getpid()
        return self._locked_store.locked_keys.get(pid) or set()

    def _add_locked_key(self, key):
        """Adds to Process Specific set of locked keys a given key"""
        pid = os.getpid()
        if pid not in self._locked_store.locked_keys:
            self._locked_store.locked_keys[pid] = set()
        self._locked_store.locked_keys[pid].add(key)

    def _remove_locked_key(self, key):
        """Removes a given key from Process Specific set of locked keys"""
        pid = os.getpid()
        if pid not in self._locked_store.locked_keys:
            self._locked_store.locked_keys[pid] = set()
        self._locked_store.locked_keys[pid].remove(key)

    def _setup_locked_store(self):
        try:
            self._locked_store.locked_keys
        except AttributeError:
            # Dict of PID(int): Locked Keys(set) to avoid issues with forking memory
            self._locked_store.locked_keys = dict()


def return_when_locked(return_value_when_locked: str = 'Locked'):
    """A decorator for functions that may raise the 'Locked' exception that returns the given value instead of letting 'Locked' to bubble.

    Usage:

    >>> lock = Lock('key')
    >>> @return_when_locked('already locked')
    >>> @lock(blocking=False)
    >>> def do_something():
    >>>     return 'done'

    When calling 'do_something()' while the lock has been already acquired, 'already locked' will be returned instead of raising the 'Locked'
    exception.
    """
    def real_decorator(fn):
        """Decorator function."""
        @wraps(fn)
        def inner(*args, **kwargs):
            """Inner function."""
            try:
                return fn(*args, **kwargs)
            except Locked:
                return return_value_when_locked
        return inner
    return real_decorator



"""Tests for the ZookeeperConnectionManager class."""

from unittest import mock

from kazoo.exceptions import ConnectionClosedError

from django.test import TestCase

from zookeeper_locks.connection import zookeeper_connection_manager

from .utils import FakeZookeeperClient


class ZookeeperConnectionManagerTestCase(TestCase):

    """Test the ZookeeperConnectionManager."""

    def test_context_manager(self):
        """Test that the context manager interface of the manager has correct is_managed property."""
        self.assertFalse(zookeeper_connection_manager.is_managed)
        with zookeeper_connection_manager:
            self.assertTrue(zookeeper_connection_manager.is_managed)
        self.assertFalse(zookeeper_connection_manager.is_managed)

    def test_nesting_context_manager(self):
        """Test that the is_manager property has a proper value when nesting context managers."""
        with zookeeper_connection_manager:
            with zookeeper_connection_manager:
                self.assertTrue(zookeeper_connection_manager.is_managed)
            self.assertTrue(zookeeper_connection_manager.is_managed)
        self.assertFalse(zookeeper_connection_manager.is_managed)

    def test_decorator(self):
        """Test that the decorator interface of the manager has correct is_managed property."""
        @zookeeper_connection_manager
        def some_function():
            """Just assert that the is_manager property is True."""
            self.assertTrue(zookeeper_connection_manager.is_managed)
        self.assertFalse(zookeeper_connection_manager.is_managed)
        some_function()
        self.assertFalse(zookeeper_connection_manager.is_managed)

    def test_getting_client_out_of_context(self):
        """Test that getting a client out of the manager context raises an AssertionError."""
        self.assertFalse(zookeeper_connection_manager.is_managed)
        with self.assertRaises(AssertionError) as cm:
            zookeeper_connection_manager.get_client()
        self.assertFalse(zookeeper_connection_manager.has_client)
        self.assertTupleEqual(
            cm.exception.args,
            ('Use the zookeeper_locks.connection.ZookeeperConnectionManager as a context manager or decorator.', )
        )

    def test_getting_client(self):
        """Test that using the connection manager creates and connects the ZookeeperClient when necessary."""
        fake_client = FakeZookeeperClient()
        with mock.patch('zookeeper_locks.locks.zookeeper_connection_manager._create_client', return_value=fake_client), \
                mock.patch.object(fake_client, 'start', wraps=fake_client.start) as start_mock:
            with zookeeper_connection_manager:
                client = zookeeper_connection_manager.get_client()
                self.assertIs(client, fake_client)
                self.assertTrue(zookeeper_connection_manager.has_client)
                self.assertTrue(fake_client.started)
                zookeeper_connection_manager.get_client()
                with zookeeper_connection_manager:
                    self.assertTrue(zookeeper_connection_manager.has_client)
                    self.assertTrue(fake_client.started)
                    zookeeper_connection_manager.get_client()
                self.assertTrue(fake_client.started)
                start_mock.assert_called_once()
            self.assertFalse(fake_client.started)
            with zookeeper_connection_manager:
                zookeeper_connection_manager.get_client()
            start_mock.assert_has_calls([mock.call(), mock.call()])

    def test_restarting_client(self):
        """Test that connection is restarted on ConnectionClosedError when it's needed."""
        fake_client = FakeZookeeperClient()
        with mock.patch('zookeeper_locks.locks.zookeeper_connection_manager._create_client', return_value=fake_client), \
                mock.patch.object(fake_client, 'restart', wraps=fake_client.restart) as restart_mock:
            with self.assertRaises(ConnectionClosedError):
                with zookeeper_connection_manager:
                    zookeeper_connection_manager.get_client()
                    raise ConnectionClosedError()
            restart_mock.assert_not_called()

            with self.assertRaises(ConnectionClosedError):
                with zookeeper_connection_manager:
                    with zookeeper_connection_manager:
                        raise ConnectionClosedError()
            restart_mock.assert_not_called()

            with self.assertRaises(ConnectionClosedError):
                with zookeeper_connection_manager:
                    with zookeeper_connection_manager:
                        zookeeper_connection_manager.get_client()
                        raise ConnectionClosedError()
            restart_mock.assert_called_once()




"""Utils for zookeeper utils tests."""

from typing import (
    Callable,
    Iterable,
    Optional,
    Type,
    Union,
)

from django.test import TestCase


class FakeZookeeperClient:

    """A dummy zookeeper client for testing purposes."""

    class Lock:

        """A dummy lock class."""

        def __init__(self, key):  # pylint: disable=unused-argument
            """Mock the original `__init__` method."""

        def acquire(self, blocking=True, timeout=None):  # pylint: disable=unused-argument
            """Mock the original `acquire` method."""
            return True

        def release(self):
            """Mock the original `release` method."""

    def __init__(self):
        """Initialize the dummy client attributes."""
        self.started = False
        self.ever_started = False
        self.restarted = False

    def start(self):
        """Set the started attribute to True."""
        self.started = True
        self.ever_started = True

    def stop(self):
        """Set the started attribute to False."""
        self.started = False

    def restart(self):
        """Set restarted attribute to True. Call stop and start."""
        self.restarted = True
        self.stop()
        self.start()


def params_to_kwargs(params_names: Iterable, params: Iterable) -> dict:
    """Merge iterables of names and values to dict"""
    return dict(zip(params_names, params))


def parametrize(params_names: Iterable, params: Union[Iterable, dict], custom_suffix_func: Optional[Callable]=None) -> Callable:
    """Decorator used to parametrize test method with supplied param_names.

    Params should be iterable with lists of params in the same order as in param_names or dict with those lists as values.
    In the first case tests suffixes will be generated from param iterable indexes, in the second from dict keys.
    To create custom tests suffixes supply custom_suffix_func which takes __cls argument (TestCase class) and all params of current test,
    should return string - custom suffix.
    """
    assert not any(param_name.startswith('__') for param_name in params_names), "Param names can't start with __."

    if isinstance(params, dict):
        params_list = params.items()
    else:
        params_list = enumerate(params)

    params = {k: params_to_kwargs(params_names, v) for k, v in params_list}
    suffix_func = custom_suffix_func or (lambda **kwargs: kwargs['__suffix'])

    def decorator(func: Callable) -> Callable:
        """Decorator function"""
        func.is_parametrized = True
        func.params = params
        func.suffix_func = suffix_func
        return func

    return decorator


def parametrized_test_case(cls: Type[TestCase]) -> Type[TestCase]:
    """Decorator for TestCase to enable use of parametrize decorator on its test methods"""
    def test_wrapper(test: Callable, **kwargs):
        """Create test executor"""
        def test_executor(instance: TestCase):
            """Execute test"""
            return test(instance, **kwargs)
        return test_executor

    for name, func in list(cls.__dict__.items()):
        if not getattr(func, 'is_parametrized', False):
            continue
        for suffix, params in func.params.items():
            setattr(cls, '{}__{}'.format(name, func.suffix_func(__suffix=suffix, __cls=cls, **params)), test_wrapper(func, **params))
        delattr(cls, name)
    return cls



"""Custom test runners for mocking the ZookeeperClient."""

from unittest import mock

from django.test.runner import DiscoverRunner

from zookeeper_locks.tests.utils import FakeZookeeperClient


class ZookeeperClientMockingTestRunner(DiscoverRunner):

    """Test runner that mocks the Zookeeper client created with the zookeeper_connection_manager."""

    def run_tests(self, *args, **kwargs):  # pylint: disable=arguments-differ
        """Mock the `_create_client` method while running tests."""
        with mock.patch('zookeeper_locks.connection.zookeeper_connection_manager._create_client', return_value=FakeZookeeperClient()):
            return super().run_tests(*args, **kwargs)



import os
from setuptools import setup


with open(os.path.join(os.path.dirname(__file__), 'README.md')) as readme:
    README = readme.read()

with open('requirements.txt') as requirements:
    required = requirements.read().splitlines()


def get_packages(package):
    """
    Return root package and all sub-packages.
    """
    return [
        dirpath for dirpath, dirnames, filenames in os.walk(package)
        if os.path.exists(os.path.join(dirpath, '__init__.py'))]


# allow setup.py to be run from any path
os.chdir(os.path.normpath(os.path.join(os.path.abspath(__file__), os.pardir)))

setup(
    name='django-zookeeper-locks',
    version='1.0.0',
    packages=get_packages('zookeeper_locks'),
    include_package_data=True,
    description='Utils for using Zookeeper locks through zookeeper in Django commands and views.',
    long_description=README,
    long_description_content_type='text/markdown',
    author='WeDeploy',
    author_email='github@wedeploy.com',
    install_requires=required,
    tests_require=[
        'mock',
    ],
    test_suite='testproject.runtests.run_tests',
    classifiers=[
        'Environment :: Web Environment',
        'Framework :: Django',
        'Intended Audience :: Developers',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
        'Programming Language :: Python :: 3.11',
    ],
)



"""Test cases for the migrate_with_zookeeper management command."""
from io import StringIO

from mock import patch

from django.core.exceptions import ImproperlyConfigured
from django.test import override_settings
from django.test.testcases import TestCase

from zookeeper_locks.locks import Lock
from zookeeper_locks.management.commands import migrate_with_zookeeper


@patch('sys.stdout', StringIO())
class MigrateWithZookeeperCommandTestCase(TestCase):

    """Test case for the migrate_with_zookeeper management command."""

    @override_settings(ZOOKEEPER_HOSTS=['localhost'])
    @patch('zookeeper_locks.management.commands.migrate_with_zookeeper.MigrationExecutor', return_value=None, side_effect=ImproperlyConfigured)
    @patch('zookeeper_locks.management.commands.migrate_with_zookeeper.nullcontext')
    @patch('django.core.management.commands.migrate.Command.handle', return_value=None)
    def test_using_default_lock(self, base_command_mocked, nullcontext_mocked, executor_mocked):
        command = migrate_with_zookeeper.Command()
        self.assertEqual(command.lock.key, 'migrations')
        command.handle()
        nullcontext_mocked.assert_not_called()
        base_command_mocked.assert_called_once()

    @override_settings(ZOOKEEPER_HOSTS=['localhost'])
    @patch('zookeeper_locks.management.commands.migrate_with_zookeeper.MigrationExecutor', return_value=None, side_effect=ImproperlyConfigured)
    @patch('zookeeper_locks.management.commands.migrate_with_zookeeper.nullcontext')
    @patch('django.core.management.commands.migrate.Command.handle', return_value=None)
    def test_using_provided_lock(self, base_command_mocked, nullcontext_mocked, executor_mocked):
        lock = Lock('test-lock')
        command = migrate_with_zookeeper.Command(lock=lock)
        self.assertEqual(command.lock, lock)
        command.handle()
        nullcontext_mocked.assert_not_called()
        base_command_mocked.assert_called_once()

    @patch('zookeeper_locks.management.commands.migrate_with_zookeeper.MigrationExecutor', return_value=None, side_effect=ImproperlyConfigured)
    @patch('zookeeper_locks.management.commands.migrate_with_zookeeper.nullcontext')
    @patch('django.core.management.commands.migrate.Command.handle', return_value=None)
    def test_using_nullcontext(self, base_command_mocked, nullcontext_mocked, executor_mocked):
        command = migrate_with_zookeeper.Command()
        command.handle()
        nullcontext_mocked.assert_called_once()
        base_command_mocked.assert_called_once()

    @patch('zookeeper_locks.management.commands.migrate_with_zookeeper.MigrationExecutor.migration_plan', return_value=False)
    @patch('zookeeper_locks.management.commands.migrate_with_zookeeper.Command.launched_with_defaults', return_value=True)
    @patch('django.core.management.commands.migrate.Command.handle')
    def test_without_unapplied_migrations(self, base_command_mocked, *other_mocks):
        command = migrate_with_zookeeper.Command()
        command.handle()
        base_command_mocked.assert_not_called()



"""Tests for the Lock class."""

from unittest import mock

from kazoo.exceptions import LockTimeout as ZookeeperLockTimeout

from django.core.exceptions import ImproperlyConfigured
from django.test import TestCase

from zookeeper_locks.locks import (
    Lock,
    Locked,
    LockTimeout,
    return_when_locked,
)

from .utils import (
    FakeZookeeperClient,
    parametrize,
    parametrized_test_case,
)


@parametrized_test_case
class LockTestCase(TestCase):

    """Test the Lock class."""

    def test_registering_locks(self):
        """Test registering and checking lock keys."""
        lock1 = Lock('key1')
        self.assertEqual(lock1.key, 'key1')
        self.assertIn(lock1.key, Lock.keys_registry)
        lock2 = Lock('key2')
        self.assertIn(lock1.key, Lock.keys_registry)
        self.assertIn(lock2.key, Lock.keys_registry)
        lock3 = None
        with self.assertRaises(ImproperlyConfigured) as cm:
            lock3 = Lock('key2')
        self.assertTupleEqual(cm.exception.args, ('Attempt to register the same key twice: key2', ))
        self.assertIsNone(lock3)
        self.assertIn(lock1.key, Lock.keys_registry)
        self.assertIn(lock2.key, Lock.keys_registry)
        del lock1
        self.assertNotIn('key1`', Lock.keys_registry)
        self.assertIn(lock2.key, Lock.keys_registry)

    @parametrize(
        ['key', 'key_params', 'zookeeper_key', 'blocking', 'timeout'],
        {
            'without_params': ['sample-key', {}, '/locks/django-zookeeper-test-app/sample-key', True, None],
            'with_params': ['sample-key-{param}', {'param': 21}, '/locks/django-zookeeper-test-app/sample-key-21', True, None],
            'blocking_timeout': ['sample-key', {}, '/locks/django-zookeeper-test-app/sample-key', True, 10.0],
            'not_blocking': ['sample-key', {}, '/locks/django-zookeeper-test-app/sample-key', False, None],
        }
    )
    def test_locking(self, key, key_params, zookeeper_key, blocking, timeout):
        """Test successful lock scenarios."""
        lock = Lock(key)
        fake_client = FakeZookeeperClient()
        self.assertFalse(fake_client.started)
        with mock.patch('zookeeper_locks.locks.zookeeper_connection_manager._create_client', return_value=fake_client), \
                mock.patch.object(fake_client.Lock, '__init__', return_value=None) as init_mock, \
                mock.patch.object(fake_client.Lock, 'acquire', return_value=True) as acquire_mock, \
                mock.patch.object(fake_client.Lock, 'release') as release_mock:
            with lock(blocking=blocking, timeout=timeout, **key_params):
                self.assertTrue(fake_client.started)
                init_mock.assert_called_once_with(zookeeper_key)
                acquire_mock.assert_called_once_with(blocking=blocking, timeout=timeout)
                release_mock.assert_not_called()
        release_mock.assert_called_once_with()
        self.assertFalse(fake_client.started)

    @parametrize(
        ['blocking', 'timeout', 'acquire_exception', 'expected_exception', 'exception_message'], {
            'blocking_timeout': [True, 10.0, ZookeeperLockTimeout, LockTimeout, 'Timeout occurred while trying to acquire a blocking lock on key'],
            'not_blocking': [False, None, None, Locked, 'Failed to acquire a non-blocking lock on key'],
        }
    )
    def test_locking_failed(self, blocking, timeout, acquire_exception, expected_exception, exception_message):
        """Test unsuccessful lock scenarios."""
        lock = Lock('key')
        fake_client = FakeZookeeperClient()
        self.assertFalse(fake_client.started)
        with mock.patch('zookeeper_locks.locks.zookeeper_connection_manager._create_client', return_value=fake_client), \
                mock.patch.object(fake_client.Lock, '__init__', return_value=None) as init_mock, \
                mock.patch.object(fake_client.Lock, 'acquire', return_value=False, side_effect=acquire_exception) as acquire_mock, \
                mock.patch.object(fake_client.Lock, 'release') as release_mock:
            with self.assertRaises(expected_exception) as cm:
                with lock(blocking=blocking, timeout=timeout):
                    self.fail('Should not be executed.')
            self.assertTupleEqual(cm.exception.args, (exception_message, ))
            init_mock.assert_called_once_with('/locks/django-zookeeper-test-app/key')
            acquire_mock.assert_called_once_with(blocking=blocking, timeout=timeout)
            release_mock.assert_not_called()
        self.assertTrue(fake_client.ever_started)
        self.assertFalse(fake_client.started)

    def test_nested_locks(self):
        """Test if nested locks are acquirable."""
        lock = Lock('key')
        fake_client = FakeZookeeperClient()
        assert not fake_client.started
        with mock.patch('zookeeper_locks.locks.zookeeper_connection_manager._create_client', return_value=fake_client), \
                mock.patch.object(fake_client.Lock, '__init__', return_value=None) as init_mock, \
                mock.patch.object(fake_client.Lock, 'acquire', return_value=True) as acquire_mock, \
                mock.patch.object(fake_client.Lock, 'release') as release_mock:

            with lock(blocking=False):
                with lock(blocking=False):
                    pass
            init_mock.assert_called_once_with('/locks/django-zookeeper-test-app/key')
            acquire_mock.assert_called_once_with(blocking=False, timeout=None)
            release_mock.assert_called_once()


class ReturnWhenLockedTestCase(TestCase):

    """Test of the decorator return_when_locked."""

    @staticmethod
    def get_test_method_with_lock(do_raise):
        """Get method wrapped by lock."""
        def lock(fn):
            """Fake lock"""
            def inner():
                """Call fake task or raise already locked"""
                if do_raise:
                    raise Locked()
                else:
                    return fn()
            return inner

        @return_when_locked('already locked')
        @lock
        def do_something_task():
            """A tested method."""
            return 'done'

        return do_something_task

    def test_locking(self):
        """Test successful lock scenarios."""
        method = self.get_test_method_with_lock(False)
        result = method()
        self.assertEqual(result, 'done')

    def test_locking_failed(self):
        """Test unsuccessful lock scenarios."""
        method = self.get_test_method_with_lock(True)
        result = method()
        self.assertEqual(result, 'already locked')
